{ File   : SCLUExt.dpr
  Author : Deniz Oezmen
  Created: 2005-12-02
  Changed: 2005-12-02

  Broken Sword 1 speech CLU extractor.
}
program SCLUExt;

{$APPTYPE CONSOLE}

uses
  SysUtils, StrUtils;

type
  TSpeechEntry = packed record
                   Offset,
                   Size  : Longword;
                 end;

  TCLUHeader   = packed record
                   DataOfs,
                   IndexOfs     : Longword;
                   SpeechEntries: array of TSpeechEntry
                 end;
                 
  TBuffer       = array of Byte;

var
  CLUHeader: TCLUHeader;
  CLUFile  : file;

function OpenCLUFile(var CLUFile: file; CLUName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(CLUFile, CLUName);

  {$I-}
  Reset(CLUFile, 1);
  {$I+}
  OpenCLUFile := (IOResult = 0)
end;

function ReadHeader(var CLUFile: file; var CLUHeader: TCLUHeader): Boolean;
var
  BytesRead: Longword;
begin
  Result := False;

  BlockRead(CLUFile, CLUHeader.DataOfs, 4, BytesRead);
  if (BytesRead < 4) or (Longint(CLUHeader.DataOfs) > FileSize(CLUFile)) then
    Exit;

  BlockRead(CLUFile, CLUHeader.IndexOfs, 4, BytesRead);
  if (BytesRead < 4) or (CLUHeader.IndexOfs > CLUHeader.DataOfs) then
    Exit;

  // skip unknown header data
  if FileSize(CLUFile) < Longint(CLUHeader.IndexOfs) then
    Exit;

  Seek(CLUFile, CLUHeader.IndexOfs);

  SetLength(CLUHeader.SpeechEntries,
    (CLUHeader.DataOfs - CLUHeader.IndexOfs) div 8);
  BlockRead(CLUFile, CLUHeader.SpeechEntries[0],
    (High(CLUHeader.SpeechEntries) + 1) * 8);

  Result := True
end;

function Prefix(FileName: string): string;
var
  i: Longint;
begin
  FileName := ExpandFileName(FileName);

  repeat
    i := Pos('\', FileName);
    FileName := Copy(FileName, i + 1, Length(FileName) - i)
  until i = 0;

  i := Length(FileName) + 1;
  repeat
    Dec(i)
  until (i = 0) or (FileName[i] = '.');

  if i > 0 then
    Result := Copy(FileName, 1, i - 1)
  else
    Result := FileName
end;

procedure Decompress(InBuffer: TBuffer; var OutBuffer: TBuffer);
var
  i,
  InPos,
  OutPos,
  InSize,
  OutSize: Longword;
  RunLen : Smallint;
begin
  InSize := High(InBuffer) + 1;

  // find data chunk
  InPos := 0;
  i := 0;
  while InPos < InSize do
  begin
    i := i shr 8 + InBuffer[InPos] shl 24;
    Inc(InPos);

    if i = $61746164 then
      Break
  end;

  if InPos < InSize then
    OutSize := InBuffer[InPos] + InBuffer[InPos + 1] shl 8 +
      InBuffer[InPos + 2] shl 16 + InBuffer[InPos + 3] shl 24
  else
  begin
    SetLength(OutBuffer, 0);
    Exit
  end;

  Inc(InPos, 4);

  SetLength(OutBuffer, OutSize + InPos);

  // should be "RIFF" ...
  for i := 0 to 3 do
    OutBuffer[i] := InBuffer[i];

  // fix RIFF chunk size (why is this wrong?)
  Inc(OutSize, InPos - 8);
  OutBuffer[4] := OutSize and $000000ff;
  OutBuffer[5] := (OutSize and $0000ff00) shr 8;
  OutBuffer[6] := (OutSize and $00ff0000) shr 16;
  OutBuffer[7] := (OutSize and $ff000000) shr 24;

  // copy rest of header
  for i := 8 to InPos + 3 do
    OutBuffer[i] := InBuffer[i];

  OutPos := InPos;

  while InPos < InSize do
  begin
    RunLen := InBuffer[InPos] + InBuffer[InPos + 1] shl 8;
    Inc(InPos, 2);

    if RunLen < 0 then
    begin
      for i := 1 to -RunLen do
      begin
        OutBuffer[OutPos] := InBuffer[InPos];
        OutBuffer[OutPos + 1] := InBuffer[InPos + 1];
        Inc(OutPos, 2)
      end;

      Inc(InPos, 2)
    end
    else
    begin
      for i := 1 to RunLen do
      begin
        OutBuffer[OutPos] := InBuffer[InPos];
        OutBuffer[OutPos + 1] := InBuffer[InPos + 1];
        Inc(OutPos, 2);
        Inc(InPos, 2)
      end
    end
  end;
end;

procedure ExtractFiles(CLUName: string);
var
  i        : Longint;
  OutFile  : file;
  Name     : string;
  InBuffer,
  OutBuffer: TBuffer;
begin
  for i := 0 to High(CLUHeader.SpeechEntries) do
    with CLUHeader.SpeechEntries[i] do
    begin
      if Size = 0 then
        Continue;

      Write('Extracting @', IntToHex(Offset, 8), ' (', Size, ' --> ');

      Name := IntToHex(Offset, 8) + '.wav';

      Seek(CLUFile, CLUHeader.DataOfs + Offset);
      SetLength(InBuffer, Size);
      BlockRead(CLUFile, InBuffer[0], Size);

      Decompress(InBuffer, OutBuffer);

      Finalize(InBuffer);

      WriteLn(High(OutBuffer) + 1, ' bytes)');

      FileMode := fmOpenReadWrite;
      Assign(OutFile, Name);
      Rewrite(OutFile, 1);

      BlockWrite(OutFile, OutBuffer[0], High(OutBuffer) + 1);

      Finalize(OutBuffer);

      CloseFile(OutFile)
    end
end;

procedure ParseFiles(FilePara: string);
var
  DirPos,
  i        : Byte;
  FileName,
  Dir,
  CLUName  : string;
  SearchRec: TSearchRec;
  PFilePara: PChar;
begin
  GetMem(PFilePara, Length(FilePara) + 1);
  StrPCopy(PFilePara, FilePara);

  // get the directory information from FilePara
  DirPos := 0;
  for i := 1 to Length(FilePara) do
    if FilePara[i] in [':', '\'] then
      DirPos := i;

  Dir := Copy(FilePara, 1, DirPos);

  FileMode := fmOpenRead;

  // search for all files matching FilePara
  if FindFirst(PFilePara, faAnyFile - faDirectory - faVolumeID,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      CLUName := Dir + FileName;

      // if this file is accessible, analyze it
      if OpenCLUFile(CLUFile, CLUName) then
      begin
        if ReadHeader(CLUFile, CLUHeader) then
        begin
          Writeln('Extracting from ', CLUName, ' ...');
          ExtractFiles(CLUName);
          Close(CLUFile)
        end
        else
          WriteLn(CLUName, ' is not a valid speech CLU file.')
      end
      else
        Writeln(CLUName, ' not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    Writeln(FilePara, ' not found')
end;

begin
  if ParamCount = 1 then
    ParseFiles(ParamStr(1))
  else
    Writeln('Incorrect number of parameters.')
end.

